

from zope.app.pagetemplate import ViewPageTemplateFile
from zope.formlib import namedtemplate


FormTemplate = namedtemplate.NamedTemplateImplementation(
    ViewPageTemplateFile('templates/form.pt')
    )

SubFormTemplate = namedtemplate.NamedTemplateImplementation(
    ViewPageTemplateFile('templates/subform.pt')
    )



"""

$Id: container.py 325 2008-07-31 03:28:51Z kapilt $
"""
from zope import schema
from zope.publisher.browser import BrowserView
from zope.formlib import form
from zope.security import proxy
from zc.table import column, table
from zope.app.pagetemplate import ViewPageTemplateFile

import simplejson, datetime

from ore.alchemist.container import stringKey
from ore.alchemist.model import queryModelInterface
from i18n import _

import core

class Getter( object ):

    def __init__(self, getter):
        self.getter = getter

    def __call__( self, item, formatter):
        return self.getter( item )

def viewLink( item, formatter ):
    return u'<a class="button-link" href="%s">View</a>'%( stringKey( item ) )

def editLink( item, formatter ):
    return u'<a class="button-link" href="%s/edit">Edit</a>'%( stringKey( item ) )

def viewEditLinks( item, formatter ):
    return u'%s %s'%(viewLink( item, formatter), editLink( item, formatter ) )


class ContainerListing( form.DisplayForm ):
    
    form_fields = form.Fields()
    mode = "listing"

    index = ViewPageTemplateFile('templates/generic-container.pt')
    
    def update( self ):
        context = proxy.removeSecurityProxy( self.context )
        columns = core.setUpColumns( context.domain_model )
        columns.append(
            column.GetterColumn( title = _(u'Actions'),
                                 getter = viewEditLinks )
            )
        self.columns = columns
        
        super( ContainerListing, self).update()
        
    def render( self ):
        return self.index()
        
    def listing( self ):
        return self.formatter()
    
    @property
    def formatter( self ):
        context = proxy.removeSecurityProxy( self.context )
        
        formatter = table.AlternatingRowFormatter( context,
                                                   self.request,
                                                   context.values(),
                                                   prefix="form",
                                                   columns = self.columns )
        formatter.cssClasses['table'] = 'listing'
        formatter.table_id = "datacontents"
        return formatter
        

    @property
    def form_name( self ):
        return "%s %s"%( self.context.domain_model.__name__, self.mode.title())

    @form.action(_(u"Add") )
    def handle_add( self, action, data ):
        self.request.response.redirect('add')


class ContainerJSONListing( BrowserView ):
    """
    paging, batching, sorting, json contents of a container
    """

    def getSort( self ):
        """ server side sort,
        @web_parameter sort - request variable for sort column
        @web_parameter dir  - direction of the sort, only once acceptable value 'desc'
        """
        sort_key, sort_dir = self.request.get('sort'), self.request.get('dir')
        domain_model = proxy.removeSecurityProxy( self.context.domain_model )
        # get sort in sqlalchemy form
        if sort_key and ( sort_key in domain_model.c ):
            column = domain_model.c[sort_key]
            if sort_dir == 'desc':
                return column.desc()
            return column
        return None
    
    def getOffsets( self ):
        nodes = []
        start, limit = self.request.get('start',0), self.request.get('limit', 25)
        try:
            start, limit = int( start ), int( limit )
            if not limit:
                limit = 100
        except ValueError:
            start, limit = 0, 100
        return start, limit 

    def getBatch( self, start, limit, order_by=None):

        # fetch the nodes from the container
        nodes = self.context.batch( offset=start,
                                    limit=limit,
                                    order_by=order_by )

        fields = list( getFields( self.context )  )
        batch = self._jsonValues( nodes, fields )
        return batch

    def _jsonValues( self, nodes, fields ):
        """
        filter values from the nodes to respresent in json, currently
        that means some footwork around, probably better as another
        set of adapters.
        """
        values = []
        for n in nodes:
            d = {}
            # field to dictionaries
            for field in fields:
                f = field.__name__
                d[ f ] = v = field.query( n )
                if isinstance( v, datetime.datetime ):
                    d[f] = v.strftime('%F %I:%M %p')
                elif isinstance( v, datetime.date ):
                    d[f] = v.strftime('%F')
                    
            values.append( d )
        return values
        
    def __call__( self ):
        start, limit = self.getOffsets( )
        sort_clause = self.getSort()
        batch = self.getBatch( start, limit, sort_clause )
        set_size = len( self.context )
        data = dict( length=set_size,
                     start=start,
                     recordsReturned=len(batch),
                     sort = self.request.get('sort'),
                     dir  = self.request.get('dir', "asc"),
                     nodes=batch )
        return simplejson.dumps( data )


def getFields( context ):
    domain_model = proxy.removeSecurityProxy( context.domain_model )
    domain_interface = queryModelInterface( domain_model )
    field_names = schema.getFieldNamesInOrder( domain_interface )
    for f in field_names:
        field = domain_interface[f]
        if isinstance( field,  ( schema.Choice, schema.Object,
                                 schema.List, schema.Tuple, schema.Bytes ) ):
            continue
        yield field



"""
Generic Content Views

"""

from zope import interface
from zope.event import notify
from zope.formlib import form
from zope.lifecycleevent import ObjectCreatedEvent
from zope.security.proxy import removeSecurityProxy
from zope.traversing.browser import absoluteURL
from zope.publisher.browser import BrowserView

from zope.app.pagetemplate import ViewPageTemplateFile

from ore.alchemist import Session
from sqlalchemy import orm

from i18n import _
import generic, core


class AddFormBase( object ):

    _next_url = None
    adapters = None

    def createAndAdd( self, data ):

        domain_model = removeSecurityProxy( self.context.domain_model )
        # create the object, inspect data for constructor args      
        try:  
            ob = generic.createInstance( domain_model, data )
        except TypeError:
            ob = domain_model()
        
        # apply any context values
        self.finishConstruction( ob )
        
        # apply extra form values
        form.applyChanges( ob, self.form_fields, data, self.adapters )
        
        # save the object, id is generated by db on flush
        self.context[''] = ob
        
        # flush so we have database id
        session = Session()
        session.flush()
        
        # fire an object created event
        notify(ObjectCreatedEvent(ob))
        
        # signal to add form machinery to go to next url
        self._finished_add = True
        
        mapper = orm.object_mapper( ob )
        
        # TODO single primary key ( need changes to base container)
        oid = mapper.primary_key_from_instance( ob )
        
        # retrieve the object with location and security information
        return self.context[ oid ]
        
    def finishConstruction(self, ob ):
        """ no op, subclass to provide additional initialization behavior"""
        return 
        
    def nextURL( self ):
        if not self._next_url:
            return absoluteURL( self.context, self.request )
        return self._next_url
        
    def update( self ):
        self.status = self.request.get('portal_status_message','')
        super( AddFormBase, self).update()

    def validateAdd(self, action, data ):
        errors = self.validateUnique(action, data )
        return errors
        
    @form.action(_(u"Cancel"), validator=core.null_validator )
    def handle_cancel( self, action, data ):
        url = self.nextURL()
        return self.request.response.redirect( url )
        
    @form.action(_(u"Save and continue editing"), condition=form.haveInputWidgets, validator='validateAdd')
    def handle_add_edit( self, action, data ):
        ob = self.createAndAdd( data )
        name = self.context.domain_model.__name__
        self._next_url = absoluteURL( ob, self.request ) + "/@@edit?portal_status_message=%s Added"%name
        
    @form.action(_(u"Save and add another"), condition=form.haveInputWidgets)
    def handle_add_and_another(self, action, data ):
        self.createAndAdd( data )
        name = self.context.domain_model.__name__
        self._next_url = absoluteURL( self.context, self.request ) + '/@@add?portal_status_message=%s Added'%name
        
    def invariantErrors( self ):        
        errors = []
        for error in self.errors:
            if isinstance( error, interface.Invalid ):
                errors.append( error )
        return errors

class Add( AddFormBase, form.AddForm ):
    """
    static add form for db content
    """

class DynamicAdd( core.DynamicFields, Add):
    """
    generic add form ( dynamic fields ) for db content
    """
    mode = "add"

ContentAddForm = DynamicAdd

class Display( BrowserView ):
    """
    Content Display
    """
    template = ViewPageTemplateFile('templates/content-view.pt')
    form_name = _("View")    
    
    def __call__( self ):
        return self.template()

ContentDisplayForm = Display
        
class ContentEditForm( BrowserView ):
    """
    Content Edit View
    """
    template = ViewPageTemplateFile('templates/content-edit.pt')
    form_name = _("Edit")
    
    def __call__( self ):
        return self.template()

class EditForm( form.EditForm ):

    adapters = None

    def setUpWidgets( self, ignore_request=False):
        self.adapters = self.adapters or {}
        self.widgets = form.setUpEditWidgets(
            self.form_fields, self.prefix, self.context, self.request,
            adapters=self.adapters, ignore_request=ignore_request
            )
            
    @form.action(_(u"Cancel"), condition=form.haveInputWidgets, validator=core.null_validator)
    def handle_cancel_action( self, action, data ):
        return core.handle_edit_action( self, action, data )            
        
    @form.action(_(u"Save"), condition=form.haveInputWidgets)
    def handle_edit_action( self, action, data ):
        return core.handle_edit_action( self, action, data )
    
    def invariantErrors( self ):        
        errors = []
        for error in self.errors:
            if isinstance( error, interface.Invalid ):
                errors.append( error )
        return errors



"""
$Id: $

based on zope.publisher.mapply adapted for class instantiation
and sa initialization hooks
"""

_marker = object()

wrapper_type = type(object.__init__)

def createInstance( klass, data ):
    """
    generically create an instance of a sqlalchemy object
    with a dictionary of data, and then return the instance
    and the remainder of the data. uses default args if found
    and no value for arg present in data.
    """
    func = klass.__init__.im_func
    func = func.func_dict.get('_oldinit', func)
    
    if isinstance( func, wrapper_type ):
        defaults = ()
        names = ()
    else:
        defaults = func.func_defaults
        names = func.func_code.co_varnames[1:]
    args = []
    used = []

    nrequired = len(names)
    if defaults:
        nrequired -= len( defaults )
    
    for i in range(len(names)):
        n = names[i]
        v = data.get(n, _marker)
        
        if v is _marker:
            if i < nrequired:
                raise TypeError("missing argument %s"%n)
            else:
                v = defaults[ i-nrequired ]
        else:
            used.append(n)
        args.append( v )
        
    for n in used:
        del data[n]

    ob = klass(*args)

    return ob
    

    
    



"""
$Id:$

Base class and functions for encaspulating dynamic fields,
form styling, and form name.

"""
import time, copy, math, itertools
from zope import interface, schema
from zope.formlib import form
from zope.formlib.namedtemplate import NamedTemplate
from zc.table import column

from sqlalchemy import orm
from sqlalchemy.orm import properties
from ore.alchemist import model, Session
from ore.alchemist.interfaces import IAlchemistContent

import pytz, datetime
import zope.event
from zope import security
from zope.security.proxy import removeSecurityProxy
from zope.lifecycleevent import Attributes
from zope.lifecycleevent import ObjectModifiedEvent

from zope.interface.common import idatetime
from zope.i18nmessageid import MessageFactory

_ = MessageFactory('zope')

class Getter( object ):
    def __init__(self, getter):
        self.getter = getter

    def __call__( self, item, formatter):
        return self.getter( item )
        
class DateGetter( object ):
    def __init__( self, getter, format="%d %B %Y %r" ):
        self.getter = getter
        self.format = format
    def __call__( self, item, formatter):
        value = self.getter( item )
        if not value:
            return "N/A"
        return value.strftime( self.format )

def filterFields( context, fields, mode ):
    """
    filter form fields, context should be an instance of the class.

    field mode should be in set of ('add', 'edit', 'search', 'view')
    
    There's a different set of use cases around field filtering,
    which involves editable, then readable, then omit based on
    permissions, but its a more involved as a use case with formlib
    """
    unwrapped = removeSecurityProxy( context )
    descriptor = model.queryModelDescriptor( unwrapped )

    check = mode in ('add', 'edit') \
            and security.canWrite or security.canAccess
    
    omit_names = []
    
    for f in fields:
        if check( context, f.__name__ ):
            continue
        omit_names.append( f.__name__)

    return fields.omit( *omit_names )
    
def setUpFields( domain_model, mode ):
    """
    setup form fields for add/edit/view/search modes, with custom widgets
    enabled from model descriptor. this expects the domain model and mode
    passed in and will return a form.Fields instance
    """
    
    domain_model = removeSecurityProxy( domain_model )
    t = time.time()

    domain_interface = model.queryModelInterface( domain_model )
    domain_annotation = model.queryModelDescriptor( domain_interface )

    search_mode = mode == 'search'

    if not domain_annotation:
        if search_mode:
            form_fields = form.Fields( *setUpSearchFields( domain_interface ) )
        else:
            form_fields = form.Fields( domain_interface )
        return form_fields

    fields = []
    columns = getattr( domain_annotation, '%s_columns'%mode )
    

    
    for field_info in columns:
        if not field_info.name in domain_interface:
            #print "bad field", field_info.name, domain_interface.__name__
            continue
        custom_widget = getattr( field_info, "%s_widget"%mode )
        if search_mode:
            fields.append(
                form.Field( setUpSearchField( domain_interface[ field_info.name ] ),
                            custom_widget = custom_widget )
                )
        else:
            fields.append(
                form.Field( domain_interface[ field_info.name ],
                            custom_widget = custom_widget )
                )
    form_fields = form.Fields( *fields )
    #print "field setup cost", time.time()-t    
    return form_fields


def setWidgetErrors( widgets, errors ):
    for widget in widgets:
        name = widget.context.getName()
        for error in errors:
            if isinstance(error, interface.Invalid) and name in error.args[1:]:
                if widget._error is None:
                    widget._error = error

def setUpColumns( domain_model ):
    """
    use model descriptor on domain model extract columns for table listings
    """
    columns = []
    domain_interface = model.queryModelInterface( domain_model )

    if not domain_interface:
        raise SyntaxError("Model must have domain interface %r"%(domain_model ) )

    domain_annotation = model.queryModelDescriptor( domain_interface )
    
    field_column_names = domain_annotation and domain_annotation.listing_columns \
                         or schema.getFieldNamesInOrder( domain_interface )

    # quick hack for now, dates are last in listings
    remainder = []
    
    for field_name in field_column_names:
        if not field_name in domain_interface:
            # we can specify additional columns for tables that are not present in the
            # the interface, iff they are fully spec'd as columns in the descriptor/annotation
            if domain_annotation \
                and field_name in domain_annotation \
                    and domain_annotation[ field_name ].listing_column:
                pass
            else:
                #print "bad field, container", field_name, domain_interface.__name__
                continue
            
        info = domain_annotation and domain_annotation.get( field_name ) or None
        
        if info is not None and info.listing_column:
            columns.append(
                info.listing_column
                )
            continue

        field = domain_interface[ field_name ]
        
        if isinstance( field, schema.Datetime ):
            remainder.append( 
                column.GetterColumn( title=field.title or field.__name__,
                                     getter=DateGetter( field.query ) )
                )
            continue
        columns.append(
            column.GetterColumn( title= ( field.title or field.__name__ ),
                                 getter = Getter( field.query ) )
            )
    columns.extend( remainder )
    return columns

def setUpSearchFields( iface ):
    " search fields shouldn't be required "
    fields = []
    for name, field in schema.getFieldsInOrder( iface ):
        if field.required:
            field = copy.deepcopy( field )
            field.required = False
        fields.append( field )
    return fields

def setUpSearchField( field ):
    " search fields shouldn't be required "    
    if not field.required:
        return field
    field = copy.deepcopy( field )
    field.required = False
    return field

def constructQuery( form_fields, domain_class, data ):
    d = {}
    for form_field in form_fields:
        field = form_field.field
        v = data.get( field.__name__ )
        if not v:
            continue
        if isinstance( field, (schema.Text, schema.ASCII)):
            # automatically convert text/schema fields into wildcards
            v = getattr(domain_class, field.__name__).like('%'+v+'%')
        else:
            v = getattr( domain_class, field.__name__) == v
        d[ field.__name__ ] = v
    return d

def getSelected( selection_column, request ):
    """
    the formlib selection column implementation wants us to pass a value
    set.. lame. we manually poke at it to get ids and dereference.

    these pair of functions assume single column integer primary keys
    """
    keys = [ k[len(selection_column.prefix):].decode('base64') \
             for k in request.form.keys() \
             if k.startswith( selection_column.prefix )  \
             and request.form.get(k,'') == 'on']
    return map( int, keys )    

def getSelectedObjects( selection_column, request, domain_model ):
    keys = getSelected( selection_column, request)
    session = Session()
    return filter( None, [ session.get( domain_model, k) for k in keys] )
    
def null_validator( form, action, data ):
    return ()
    
class BaseForm( form.FormBase ):
    
    template = NamedTemplate('alchemist.form')        
    
    def invariantErrors( self ):
        errors = []
        for error in self.errors:
            if isinstance( error, interface.Invalid ):
                errors.append( error )
        return errors


def columns( mapper ):
    for p in mapper.iterate_properties:
        if isinstance( p, properties.ColumnProperty ):
            yield p

def unique_columns( mapper ):
    for cp in columns( mapper ):
        for c in cp.columns:
            if c.unique:
                yield cp.key, c
            
class DynamicFields( object ):

    mode = None # required string attribute
    template = NamedTemplate('alchemist.form')
    form_fields = form.Fields()

    def getDomainModel( self ):
        return getattr( self.context, 'domain_model', self.context.__class__)
    
    def update( self ):
        """
        override update method to setup fields dynamically before widgets are
        setup and actions called.
        """
        domain_model = self.getDomainModel()
        self.form_fields = setUpFields( domain_model, self.mode )
        super( DynamicFields, self).update()
        setWidgetErrors( self.widgets, self.errors )
        
    @property
    def form_name( self ):
        # play nice w/ containers or content views, or content
        domain_model = getattr( self.context, 'domain_model', None)
        if domain_model is None:
            domain_model = getattr( self, 'domain_model', None)
            if domain_model is None and IAlchemistContent.providedBy( self.context ):
                domain_model = self.context.__class__
        if domain_model is None:
            return self.mode.title()
        return "%s %s"%(self.mode.title(), domain_model.__name__)

    def splitwidgets( self ):
        """
        return two lists of form fields split into two, int overflow into
        first column.
        """
        # give first and second set of field names
        field_count = len( self.form_fields )
        first_half_count = math.ceil( field_count / 2.0 )
        first_half_widgets, second_half_widgets = [], []
        
        for idx, field in itertools.izip( itertools.count(), self.form_fields ):
            if idx < first_half_count:
                first_half_widgets.append( self.widgets[ field.__name__ ] )
            else:
                second_half_widgets.append( self.widgets[ field.__name__ ] )

        return first_half_widgets, second_half_widgets

    def validateUnique( self, action, data ):
        """
        verification method for adding or editing against anything existing
        that conflicts with an existing row, as regards unique column constraints.
        
        this is somewhat costly as we basically reinvoke parsing form data twice,
        we need to extend zope.formlib to ignore existing values in the data dictionary
        to avoid this.
        
        its also costly because we're doing a query per unique column data we find, in
        order to unambigously identify the field thats problematic. but presumably those
        are very fast due to the unique constraint.
        
        TODO : assumes column == attribute         
        """
        errors = form.getWidgetsData( self.widgets, self.prefix, data ) + \
                 form.checkInvariants(self.form_fields, data) 
        if errors:
            return errors
        
        domain_model = removeSecurityProxy( self.getDomainModel() )
        
        # find unique columns in data model.. TODO do this statically
        mapper = orm.class_mapper( domain_model  )
        ucols = list( unique_columns( mapper ) )

        errors = []
        # query out any existing values with the same unique values,
        
        s = Session()        
        # find data matching unique columns
        for key, col in ucols:
            if key in data:
                # on edit ignore new value if its the same as the previous value
                if isinstance( self.context, domain_model ) \
                   and data[key] == getattr( self.context, key, None):
                   continue
                
                value = s.query( domain_model ).filter( col == data[key] ).count()
                if not value:
                    continue
                
                widget = self.widgets[ key ]
                error = form.WidgetInputError( widget.name, 
                                               widget.label, 
                                              u"Duplicate Value for Unique Field" )
                widget._error = error
                errors.append( error )
        
        if errors:
            return tuple(errors)

# formlib 3.5.0 backports.. these variants will send field descriptions on edit

def handle_edit_action(self, action, data):
    descriptions = applyData(self.context, self.form_fields, data,
                             self.adapters)
    if descriptions:
        descriptions = [Attributes(interface, *tuple(keys))
                        for interface, keys in descriptions.items()]
        zope.event.notify(ObjectModifiedEvent(self.context, *descriptions))
        formatter = self.request.locale.dates.getFormatter(
            'dateTime', 'medium')
        try:
            time_zone = idatetime.ITZInfo(self.request)
        except TypeError:
            time_zone = pytz.UTC
            
        status = _("Updated on ${date_time}",
                   mapping={ 'date_time':
                             formatter.format( datetime.datetime.now(time_zone) ) }
                   )
        self.status = status
    else:
        self.status = _('No changes')
        
def applyData(context, form_fields, data, adapters=None):
    if adapters is None:
        adapters = {}

    descriptions = {}

    for form_field in form_fields:
        field = form_field.field
        # Adapt context, if necessary
        interface = form_field.interface
        adapter = adapters.get(interface)
        if adapter is None:
            if interface is None:
                adapter = context
            else:
                adapter = interface(context)
            adapters[interface] = adapter

        name = form_field.__name__
        newvalue = data.get(name, form_field) # using form_field as marker
        if (newvalue is not form_field) and (field.get(adapter) != newvalue):
            descriptions.setdefault(interface, []).append(field.__name__)
            field.set(adapter, newvalue)

    return descriptions            




from zope.viewlet.interfaces import IViewletManager
from zope import interface
from zope.interface.common.sequence import ISequence

class IAlchemistLayer( interface.Interface ):
    """ Alchemist UI Layer """

class IContentEditManager( IViewletManager ):
    """ viewlet manager interface """

class IContentViewManager( IViewletManager ):
    """ viewlet manager interface """    

class IIterableSequence( ISequence ):
    def __iter__(  ):
        """iterate through elements of a sequence
        """




from zope import interface
from zope.viewlet import viewlet, manager
from zope.formlib import form
from zope.formlib.namedtemplate import NamedTemplate

from i18n import _
import core
import content

class FormViewlet( form.SubPageForm, viewlet.ViewletBase ):

    __init__ = viewlet.ViewletBase.__init__

class AddFormViewlet( content.AddFormBase, FormViewlet ):
    """
    add form viewlet
    """
    __init__ = FormViewlet.__init__

class EditFormViewlet( form.SubPageEditForm, viewlet.ViewletBase ):

    __init__ = viewlet.ViewletBase.__init__

    adapters = None

    def setUpWidgets( self, ignore_request=False):
        self.adapters = self.adapters or {}
        self.widgets = form.setUpEditWidgets(
            self.form_fields, self.prefix, self.context, self.request,
            adapters=self.adapters, ignore_request=ignore_request
            )
            
    @form.action(_(u"Cancel"), condition=form.haveInputWidgets, validator=core.null_validator)
    def handle_cancel_action( self, action, data ):
        return core.handle_edit_action( self, action, data )            
        
    @form.action(_(u"Save"), condition=form.haveInputWidgets)
    def handle_edit_action( self, action, data ):
        return core.handle_edit_action( self, action, data )
    
    def invariantErrors( self ):        
        errors = []
        for error in self.errors:
            if isinstance( error, interface.Invalid ):
                errors.append( error )
        return errors
            
class DisplayFormViewlet( form.SubPageDisplayForm, viewlet.ViewletBase ):
    
    __init__ = viewlet.ViewletBase.__init__

class AttributesEditViewlet( core.DynamicFields, EditFormViewlet ):

    mode = "edit"
    template = NamedTemplate('alchemist.subform')
    form_name = _(u"General")
    
class AttributesViewViewlet( core.DynamicFields, DisplayFormViewlet ):

    mode = "view"
    template = NamedTemplate('alchemist.subform')    
    form_name = _(u"General")

class ContentViewletManager( manager.ViewletManagerBase ):
    
    def sort( self, viewlets ):
        sorted( viewlets )
        #viewlets.sort( sorter )
        return viewlets



"""
$Id: $

UI/Viewlets for Managing Relations 

"""

from zope import interface
from zope.formlib import form, namedtemplate
from zope.app.pagetemplate import ViewPageTemplateFile
from zope.security import proxy
from zc.table import table, column
from ore.alchemist import Session


import sqlalchemy as rdb
from sqlalchemy import orm
import viewlet, core

class RelationTableBaseViewlet( viewlet.FormViewlet ):

    domain_model  = None # domain model of the other end of this relationship
    property_name = None # property on context's domain model which denotes the relationship

    selection_column = column.SelectionColumn( lambda item: str(item.id), name="selection")
    
    results = () # any subclass is expected to populate as it will.
    form_fields = form.Fields()  # these as well

    @property
    def form_name( self ):
        n = self.property_name.title()
        return n.replace('_', ' ')

    @property
    def prefix( self ):
        return self.property_name

    def setUpColumns( self ):
        columns = core.setUpColumns( self.domain_model )
        return columns

    def setUpFormatter( self ):
        """ setup table listing formatter """
        columns = self.setUpColumns()

        formatter = table.StandaloneFullFormatter(
            self.context,
            self.request,
            self.results,
            prefix=self.prefix,
            columns = columns
            )

        formatter.cssClasses['table'] = 'yui-dt-table'
        return formatter
    
    def render_listing( self ):
        """
        render a listing of all the currently associated objects or
        search results.
        """
        formatter = self.setUpFormatter()
        return formatter()

class RelationStackedBaseViewlet( viewlet.FormViewlet ):
    """
    display viewlet for stacked objects
    """

class One2OneBase( core.DynamicFields ):

    template = namedtemplate.NamedTemplate('alchemist.subform')
    
    def getDomainModel( self ):
        return self.domain_model

    def getValue( self ):
        value = getattr( self.context, self.property_name, None )
        if value is None:
            # create one on the fly
            return self.domain_model()
        return value

    @property
    def prefix( self ):
        return self.property_name    

    @property
    def model_schema( self ):
        return list( interface.implementedBy(self.domain_model) )[0]

    
class One2OneDisplay( One2OneBase, viewlet.DisplayFormViewlet ):    
    form_name = "one2one display"
    mode = 'view'
    actions = ()

    notset_template = ViewPageTemplateFile('templates/one2one-unset.pt')

    def setUpWidgets( self, ignore_request=False ):
        self.adapters = { self.model_schema : self.getValue() }
        self.widgets = form.setUpEditWidgets(
            self.form_fields, self.prefix, self.context, self.request,
            adapters=self.adapters, for_display=True,
            ignore_request=ignore_request
            )

    def render( self ):
        if self.getValue() != self.getValue():
            return self.notset_template()
        return super( One2OneDisplay, self).render()
    
class One2OneEdit( One2OneBase, viewlet.EditFormViewlet ):

    form_name = 'one2one edit'
    mode = 'edit'

    def setUpWidgets( self, ignore_request=False ):
        self.adapters = { self.model_schema : self.getValue() }
        self.widgets = form.setUpEditWidgets(
            self.form_fields, self.prefix, self.context, self.request,
            adapters=self.adapters, ignore_request=ignore_request
            )

    @form.action(u"Apply", condition=form.haveInputWidgets)
    def handle_edit_action( self, action, data):
        super( One2OneEdit, self).handle_edit_action.success( data )
        value = self.adapters[ self.model_schema ]
        # XXX we need better detection.. alternative is trying to use
        # something like an object modified event handler...
        if self.status.startswith('Updated'):
            # save or update if modified.
            Session().save_or_update( value )
            # associate to self
            setattr( self.context, self.property_name, value )            
                
class One2ManyDisplay( RelationStackedBaseViewlet ):
    """
    display viewlet for stacked objects (m) of one 2 many relation.
    """
        
class One2ManyEdit( RelationStackedBaseViewlet ):
    """
    edit viewlet for stacked objects (m) of one 2 many relation.    
    """

class Many2ManyDisplay( RelationTableBaseViewlet ):

    template = ViewPageTemplateFile('templates/relation-view.pt')
    actions = ()
    
    def update( self ):
        # populate results with values from relation
        self.results = getattr( self.context, self.property_name )
        super( Many2ManyDisplay, self).update()
        
class Many2ManyEdit( RelationTableBaseViewlet ):

    selection_column = column.SelectionColumn( lambda item: str(item.id), name="selection")
    state = 'listing'
    template = ViewPageTemplateFile('templates/relation-edit.pt')

    def setUpColumns( self ):
        columns = super( Many2ManyEdit, self).setUpColumns()
        columns.insert(0, self.selection_column )
        return columns
    
    def update( self ):
        # we capture state from the last processed action, so our next action
        # becomes available.
        self.state = self.request.get( "%s.state"%self.property_name, self.state )
        # our previous state when we process searches is the add state, in order
        # to process the form values for the search action, we need to have search
        # widgets setup prior to the handler invocation, so do that here..
        if self.state == 'add':
            self.form_fields = core.setUpFields( self.domain_model, mode='search')
        
        # if we're doing a listing prepopulate state before rendering so we
        # display empty results status, and for a delete do it eagerly so
        # we can check our condition
        if self.state in ( 'listing', 'delete'):
            self.results = getattr( self.context, self.property_name)

        super( Many2ManyEdit, self).update()
        
        # if our state changes to listing
        if not self.results and self.state == 'listing':
            self.results = getattr( self.context, self.property_name)
            
    def condition_search( self, action):
        return self.state == 'add'
    
    @form.action( u"Search", condition='condition_search')
    def handle_search( self, action, data ):
        """
        search the other end point of m2m relationship for rows to
        associate.
        """
        self.state = 'search'
        d = core.constructQuery( self.form_fields, self.domain_model, data )
        
        context = proxy.removeSecurityProxy( self.context )
        
        mapper = orm.class_mapper( self.domain_model )
        instance_pkey = mapper.primary_key_from_instance
        pkey = mapper.primary_key[0]
        
        query = Session().query( self.domain_model)
        query = query.filter(
            rdb.not_(
                pkey.in_(
                    [ ob.id for ob in getattr( context, self.property_name) ]
                    ) )
            )
        
        if not d:
            self.results = query.all()
            return
                  
        self.results = query.filter( *(d.values())).all()
        
    def condition_add( self, action):
        return self.state == 'listing'
        
    @form.action( u"Add", condition='condition_add')
    def handle_add( self, action, data ):
        """
        sends user to search form.
        """
        self.state = 'add'
        session = Session()
        related_count = session.query( self.domain_model).count()
        if related_count > 20:
            self.form_fields = core.setUpFields( self.domain_model, mode='search')
        else:
            self.state = 'search'            
            query = session.query( self.domain_model )
            context = proxy.removeSecurityProxy( self.context )
            
            collection = getattr( context, self.property_name)
            if collection:
                mapper = orm.class_mapper( self.domain_model )
                instance_pkey = mapper.primary_key_from_instance
                pkey = mapper.primary_key[0]
                query = query.filter(
                    rdb.not_( pkey.in_( [ instance_pkey(ob)[0] for ob in getattr( context, self.property_name) ] ) )
                    )
            self.results = query.all()
        
    def condition_delete( self, action):
        return self.state == 'listing' and self.results
    
    @form.action( u"Delete", condition='condition_delete')
    def handle_delete( self, action, data):
        """
        delete the selected items from the m2m relationship.
        """
        # reset form state
        self.state = 'listing'
        
        # first we need to dereference the selection column values into
        # objects to disassociate.
        values = core.getSelectedObjects( self.selection_column,
                                     self.request,
                                     self.domain_model )

        if not values:
            self.status = "No %s Selected"%( self.domain_model.__name__.title()+'s')
            return
        # remove objects from association
        association = getattr( self.context, self.property_name )
        for v in values:
            if v in association:
                association.remove( v )
                
        # pluralize properly
        if len(values) > 1:
            self.status = "%s Removed"%(self.domain_model.__name__.title()+'s')
        else:
            self.status = "%s Removed"%self.domain_model.__name__.title()

    def condition_associate( self, action ):
        return self.state == 'search'
        
    @form.action( u"Associate", condition='condition_associate')
    def handle_associate( self, action, data ):
        """
        associate a search result with the context.
        """
        # reset viewlet state to listing of associated objects
        self.state = 'listing'

        # first we need to dereference the selection column values into
        # objects to associate.
        values = core.getSelectedObjects( self.selection_column,
                                     self.request,
                                     self.domain_model )

        # add objects to the association
        association = proxy.removeSecurityProxy( getattr( self.context, self.property_name ) )

        for v in values:
            if v not in association:
                association.append( v )

        # pluralize properly
        if len(values) > 1:
            self.status = "%s Associated"%(self.domain_model.__name__.title()+'s')
        else:
            self.status = "%s Associated"%self.domain_model.__name__.title()

        # reset selection column
        self.form_reset = True
    
    def condition_cancel( self, action):
        return self.state != 'listing'
    
    @form.action( u"Cancel", condition='condition_cancel')
    def handle_cancel( self, action, data ):
        """
        cancel the current operation and go back to the view
        """
        self.state = 'listing'

def groupedRelationGetter( item, formatter ):
    rel_types = formatter.id_reltype_map.get( item.id, () )
    return ", ".join( rel_types )
    
class GroupedMany2Many( RelationTableBaseViewlet ):

    group_name = None # name of the group
    properties = ()   # sequence of property names
    
    @property
    def form_name( self ):
        return self.group_name

    @property
    def prefix( self ):
        return self.group_name

    def setUpColumns( self ):
        columns = core.setUpColumns( self.domain_model )
        columns.append( column.GetterColumn( title="Relation",
                                             getter = groupedRelationGetter ) )
        return columns
    
    def setUpFormatter( self ):
        formatter = super( GroupedMany2Many, self ).setUpFormatter()
        formatter.id_reltype_map = self.id_reltype_map
        return formatter

    def setUpValues( self ):
        values = []
        id_reltype_map = {}
        
        for p in self.properties:
            if p.endswith('s'):
                pn = p[:-1]
            else:
                pn = p
            pvalues = getattr( self.context, p )
            for v in pvalues:
                if v.id in id_reltype_map:
                    id_reltype_map[ v.id ].append( pn )
                else:
                    id_reltype_map[ v.id ] = [pn]
                    values.append( v )
                    
        return values, id_reltype_map
    
class GroupedMany2ManyDisplay( GroupedMany2Many ):

    template = ViewPageTemplateFile('templates/relation-view.pt')
    actions = ()
    
    def update( self ):
        self.results, self.id_reltype_map = self.setUpValues()
        super( GroupedMany2ManyDisplay, self).update()
        


        



import os
from setuptools import setup, find_packages

def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()

setup(
    name="alchemist.ui",
    version="0.4.0",
    install_requires=['setuptools',
                      'ore.alchemist',
                      'zope.formlib',
                      'zc.table',
                      'zope.viewlet',
                      'simplejson'],
    packages=find_packages(exclude=["*.tests"]),
    namespace_packages=['alchemist'],
    package_data = {
      '': ['*.txt', '*.zcml', '*.pt'],
    },
    zip_safe=False,
    author='Kapil Thangavelu',
    author_email='kapil.foss@gmail.com',
    description="user interface components for use with zope3 relational database applications, crud, listing and relation views",
    long_description = read('changes.txt'),
    license='LGPL',
    classifiers=['Programming Language :: Python',
                 'Framework :: Zope3',
                 ],        
    keywords="zope zope3",
    )




from xml.sax.saxutils import escape
from zope.schema.vocabulary import SimpleVocabulary
from zope.app.form.browser.widget import renderElement
import zope.app.form.browser as widgets

from ore.alchemist import Session

class URLDisplayWidget( widgets.URIDisplayWidget ):
    """
    default renders with empty text, causing issues
    """
    linkTarget = None

    def __call__(self):
        if self._renderedValueSet():
            content = self._data
        else:
            content = self.context.default
        if not content:
            return u''
        content = escape(content)
        kw = dict(contents=content, href=content)
        if self.linkTarget:
            kw["target"] = self.linkTarget
        return renderElement("a", **kw)    
    
class LongTextWidget( widgets.TextWidget ):

    displayWidth = 60

class YesNoDisplayWidget( widgets.DisplayWidget ):

    def __call__(self):
        if self._renderedValueSet():
            value = self._data
        else:
            value = self.context.default
        if value:
            return u"Yes"
        else:
            return u"No"

def YesNoInputWidget(field, request, true='Yes', false='No'):
    vocabulary = SimpleVocabulary.fromItems( ((true, True), (false, False)) )
    widget = widgets.RadioWidget(field, vocabulary, request)
    widget.size = 2
    widget.required = False
    return widget    

class NamedFkGetter( object ):
    def __init__( self, attribute, model ):
        self.attribute = attribute
        self.model = model
        
    def __call__( self, item, formatter ):
        value = getattr( item, self.attribute, None)
        if value is None:
            return None
        #return IModelDescriptor( object ).title
        associated = Session().get( self.model, value)
        return associated.short_name




from zope.publisher.interfaces.browser import IBrowserPublisher
from zope.publisher.interfaces import NotFound
from zope.publisher.browser import BrowserView
from zope.contentprovider.interfaces import IContentProvider
from zope.viewlet.interfaces import IViewlet
import zope.security.checker


class MethodPublisher( BrowserView, zope.location.Location ):

    zope.interface.implements( IBrowserPublisher )

    __Security_checker__ = zope.security.checker.NamesChecker((
        '__call__', 'browserDefault', 'publishTraverse'))
    
    def __init__( self, context, request, method):
        self.context = context
        self.request = request
        self.method = method

    def browserDefault( self, request ):
        return self, ()

    def publishTraverse( self, request, name ):
        # shouldn't ever be sub traversing from here
        raise NotFound( self, name, request )
    
    def __call__( self ):
        return self.method()

    def __getParent(self):
        return hasattr(self, '_parent') and self._parent or self.context

    def __setParent(self, parent):
        self._parent = parent

    __parent__ = property(__getParent, __setParent)


class ViewletTraverse( object ):
    pass

class ViewletTraversableMixin( object ):
    """
    supports ambigious traversal of formlib viewlets, ie. all viewlets
    of a given provider have are checked against the request to see if
    they were submitted, if a submitted viewlet is found it is updated
    and rendered
    """
    def publishTraverse(self, request, name):
        # could encapsulate to separate view
        value = getattr( self, name, None)
        if not callable( value ):
            raise NotFound(self, name, request)
        return MethodPublisher( self, self.request, self.viewlets ) 

    provider_name = ""
    
    def viewlets( self ):
        """
        traversable viewlets
        """
        # find the action in the request
        action = None
        for i in self.request.form.keys():
            if 'actions' in i:
                action = i
                break

        # XXX - send error status message
        if not action:
            raise NotFound("Viewlet Not Found")

        # do a dance to find the applicable viewlet
        # we need to know which viewlet managers we're using to do
        # this.
        
        # first lookup the edit manager
        viewlet_manager = zope.component.queryMultiAdapter(
            ( self.context, self.request, self ),
            IContentProvider,
            self.provider_name
            )

        # next lookup its viewlets
        viewlets = zope.component.getAdapters(
            (self.context, self.request, self, viewlet_manager),
            IViewlet
            )

        active_viewlet = None
        
        # determine the applicable viewlet based on prefix
        for name, viewlet in viewlets:
            prefix = getattr( viewlet, 'prefix', None)
            if prefix and action.startswith( prefix ):
                active_viewlet = viewlet

        if active_viewlet is None:
            raise NotFound("Viewlet Not Found")

        # render the viewlet
        active_viewlet.update()
        rendered =  active_viewlet.render()
        return rendered    
